//! \file    Connection.hpp
//! \brief   Declaration and implementation of the Connection class, which
//!          handles connections between modules.
//! \version 2008/07/08 (yyyy/mm/dd) 
//! \author  Luigi Nardi <luiginardi(at)gmail.com> 
//
// This file is part of YAO, a framework for variational assimilation in
// numerical models.
// Copyright (c) 2001-onwards LOCEAN. All rights reserved.
//
// This program may be redistributed and/or modified under the terms of the GNU
// General Public License as published by the Free Software Foundation, either
// version 2 of the license, or (at your option) any later version.
//
// This program is provided AS IS and WITHOUT ANY WARRANTY, including any
// implied warranty of DESIGN, MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.


#ifndef YAO_CONNECTION_HPP_INCLUDED
#define YAO_CONNECTION_HPP_INCLUDED

#include <cstdlib>
#include <cmath>
#include "../help/Helper.hpp"
#include "Modul.hpp"
#include "Table.hpp"
#include "Trajectory.hpp"

#include <vector>
#include <string>

//#define MAX_IO_NEIBOUR  1000
#define MAX_IO_NEIBOUR  10000

namespace yao {

  /** 
   *  Connection class, handles the connections between modules.
   *  
   *  It allows to stock all the information about the directive ctin-m of the description file.
   *  It allows also to perform the basic operations get and set of the members and some other operations on these members like
   *  test on the options of the ctin-m directive.
   *  This class has two members groupes:
   *  1) members needed in order to read the description file and manipolate this first stage of information.
   *  2) members needed in order to get and set the information already manipulated: the second stage of information.
   *  Before arrive to the second stage of information we have to perform some check and we have to translate the #, %, ~, +, - operators in
   *  simple connection. We mean that one object of the Connection class in the second stage represents one connection in the graph and not a multiple
   *  connection like we can define in the description file with the operators #, %, ....
   *  The members of the class can concern the first or the second stage. We put a mark in the member specification.
   *  If the member concern both stages we do not put anything.
   *
   *  The class has some significant property: "from", "N� output", "Ci", "Cj", "Ck", "Ct", "to", "N� input". These can be retrieved with the methods:
   *  getPropertyCount, getPropertyName and getProperty. These methods are useful above all in the second stage and they are used to shows 
   *  in a structured way the properties of the theConnectionTable object that is a table composed of Connection objects.
   *
   */
  class Connection {
  
  public:

    /**
     *	(i,j,k,t) connection. 
     *	The vector contains for the first stage:
     *	1) all the coordinate interval (i.e.: ci[0]=1 ==> i-1; ci[1]=2 ==> i-2) if the flag relative (rel) is seted;
     *	  If the connection is relative the connection is like: (i+ci, j+cj, k+ck). Moreover in the connection one
     *	  could define an interval (i.e.: i%3), so in these vectors we stock all the values. 
     *	2) the number of the coordinate if the flag is absolute;
     *	The vector contains for the second stage:
     *	just a number that identifies the relative or absolute coordinate.
     */
    vector<int> ci, cj, ck, ct;			      

    //! Default constructor.
    Connection():  inBegin(0), outBegin(0), inEnd(0), outEnd(0), reli(0), relj(0), relk(0), relt(0), t(false), T(false){}
    
    /** 
     *	Set the number of the input entry (only for the second stage). It can be only absolute.
     *	@param value the entry of the module being connected.
     */
    void setIn(int value){ inBegin = value; } 
    
    /** 
     *	Set the number of the output (only for the second stage). It can be only absolute.
     *	@param value the output of the module being connected.
     */
    void setOut(int value){ outBegin = value; } 
    
    /** 
     *	Set the number of the time coordinate (only for the second stage). It can be absolute or relative in function of relt.
     *	The concept of t and T private attributes is different. The others are seted if they are present in the description file.
     *	@param value the t coordinate of the module being connected.
     */
    void setT(int value){ ct.clear(); ct.push_back(value); }
    
    /** 
     *	Set the number of the i coordinate (only for the second stage). It can be absolute or relative in function of reli.
     *	@param value in the second stage set the i coordinate of the module being connected.
     */
    void setI(int value){ ci.clear(); ci.push_back(value); }
    
    /** 
     *	Set the number of the j coordinate (only for the second stage). It can be absolute or relative in function of relj.
     *	@param value the j coordinate of the module being connected.
     */
    void setJ(int value){ cj.clear(); cj.push_back(value); }
    
    /** 
     *	The number of the k coordinate. It can be absolute or relative in function of relk.
     *	@param value in the second stage set the k coordinate of the module being connected. 
     */
    void setK(int value){ ck.clear(); ck.push_back(value); }

    /** 
     *	Get the number of the input (only for the second stage). It can be only absolute.
     *	@return the entry of the module connected. 
     */
    int getIn(){ return inBegin; }
    
    /** 
     *	Get the number of the output (only for the second stage). It can be only absolute.
     *	@return the output of the module connected.
     */
    int getOut(){ return outBegin; }

    /** 
     *	Get the number of t coordinate (only for the second stage). It can be absolute or relative in function of relt.
     *	@return the t coordinate of the module connected.
     */
    int getT(){ return ct[0]; }

    /** 
     *	Get the number of i coordinate (only for the second stage). It can be absolute or relative in function of reli.
     *	@return the i coordinate of the module connected.
     */
    int getI(){ return ci[0]; }
    
    /** 
     *	Get the number of j coordinate (only for the second stage). It can be absolute or relative in function of relj.
     *	@return the j coordinate of the module connected.
     */
    int getJ(){ return cj[0]; }
    
    /** 
     *	Get the number of k coordinate (only for the second stage). It can be absolute or relative in function of relk.
     *	@return the k coordinate of the module connected. 
     */
    int getK(){ return ck[0]; }

    /** 
     *	Get a string that identifies the Connection. It's "outModule ===> inModule". 
     *	@return the name of the connection.
     */
    string getName(){ return outModule + " ===> " + inModule; }

    /** 
     *	Set the lower edge of an input range (only for the first stage). The range can be a simple point in function of inEnd. 
     *	The range can only be absolute.
     *	@param anInBegin the edge. 
     */
    void setInBegin(int anInBegin){ inBegin=anInBegin; }

    /** 
     *	Get the lower edge of an input range (only for the first stage). The range can be a simple point in function of inEnd. 
     *	The range can only be absolute.
     *	@return the edge. 
     */
    int getInBegin() { return inBegin; }
    
    /** 
     *	Set the upper edge of an input range (only for the first stage). If inEnd is 0 the range is a single point stocked in inBegin. 
     *	The range can only be absolute.
     *	@param anInEnd the edge. 
     */
    void setInEnd(int anInEnd){ inEnd=anInEnd; }

    /** 
     *	Get the upper edge of an input range (only for the first stage). If inEnd is 0 the range is a single point stocked in inBegin. 
     *	The range can only be absolute.
     *	@return the edge. 
     */
    int getInEnd() { return inEnd; }

    /** 
     *	Set the lower edge of an output range (only for the first stage). The range can be a simple point in function of outEnd. 
     *	The range can only be absolute.
     *	@param anOutBegin the edge. 
     */
    void setOutBegin(int anOutBegin){ outBegin=anOutBegin; }

    /** 
     *	Get the lower edge of an output range (only for the first stage). The range can be a simple point in function of outEnd. 
     *	The range can only be absolute.
     *	@return the edge. 
     */
     int getOutBegin() { return outBegin; }
    
    /** 
     *	Set the upper edge of an output range (only for the first stage). If outEnd is 0 the range is a single point stocked in outBegin. 
     *	The range can only be absolute.
     *	@param anOutEnd the edge. 
     */
    void setOutEnd(int anOutEnd){ outEnd=anOutEnd; }

    /** 
     *	Get the output module. 
     *	@return the output module. 
     */
    int getOutEnd() { return outEnd; }
    
    /** 
     *	Set the input module that is the module in the connection that receive the information from the output module. 
     *	@param anInModule the input module. 
     */
    void setInModule(string anInModule){ inModule=anInModule; }

    /** 
     *	Get the input module. 
     *	@return the input module. 
     */
    string getInModule(){ return inModule; }
    
    /** 
     *	Set the output module that is the module in the connection that give the information to the input module. 
     *	@param anOutModule the output module. 
     */
    void setOutModule(string anOutModule){ outModule=anOutModule; }

    /** 
     *	Get the output module. 
     *	@return the output module. 
     */
    string getOutModule(){ return outModule; }

    /** 
     * Get the input total number of the inModule range (only for the first stage).
     * @return the number of elements of the range.
     */
    int getInElements(){ return inEnd ? inEnd - inBegin + 1: 1; }

    /** 
     * Get the output total number of the outModule range (only for the first stage).
     * @return the number of elements of the range.
     */
    int getOutElements(){ return outEnd ? outEnd - outBegin + 1: 1; }

    /** 
     * Get the i total range number (only for the first stage).
     * @return the number of elements of the range.
     */
    int getIElements(){ return ci.size(); }

    /** 
     * Get the j total range number (only for the first stage).
     * @return the number of elements of the range.
     */
    int getJElements(){ return cj.size(); }

    /** 
     * Get the k total range number (only for the first stage).
     * @return the number of elements of the range.
     */
    int getKElements(){ return ck.size(); }

    /** 
     * Get the t total range number (only for the first stage).
     * @return the number of elements of the range.
     */
    int getTElements(){ return ct.size(); }

    /** 
     * Control if in the description file the i coordinate has been inserted in the directive ctin-m.
     * @return true if i is present in the description file.
     */
    //bool isI(){ return ci.size(); }
    bool isI(){ return reli; }

    /** 
     * Control if in the description file the j coordinate has been inserted in the directive ctin-m.
     * @return true if j is present in the description file.
     */
    //bool isJ(){ return cj.size(); }
    bool isJ(){ return relj; }

    /** 
     * Control if in the description file the k coordinate has been inserted in the directive ctin-m.
     * @return true if k is present in the description file.
     */
    //bool isK(){ return ck.size(); }
    bool isK(){ return relk; }

    /** 
     * Control if in the description file the t coordinate has been inserted in the directive ctin-m. Be careful t is not T.
     * @return true if t is present in the description file.
     */
    bool ist(){ return t; }

    /** 
     * Control if in the description file the T coordinate has been inserted in the directive ctin-m. Be careful T is not t.
     * @return true if T is present in the description file.
     */
    bool isT(){ return T; }
    
    /** 
     * Set relative or absolute displacement for i coordinate. If setted the content of ci is relative.
     * @param aReli 0 not present in the ctin-m directive, -1 absolute displacement, 1 relative displacement.
     */
    void setReli(short aReli){ reli=aReli; }
    
    /** 
     * Get relative or absolute displacement for i coordinate. 
     * @return 0 not present in the ctin-m directive, -1 absolute displacement, 1 relative displacement.
     */
    short getReli(){ return reli; }
    
    /** 
     * Set relative or absolute displacement for j coordinate. If setted the content of cj is relative.
     * @param aRelj 0 not present in the ctin-m directive, -1 absolute displacement, 1 relative displacement.
     */
    void setRelj(short aRelj){ relj=aRelj; }
    
    /** 
     * Get relative or absolute displacement for j coordinate. 
     * @return 0 not present in the ctin-m directive, -1 absolute displacement, 1 relative displacement.
     */
    short getRelj(){ return relj; }
    
    /** 
     * Set relative or absolute displacement for k coordinate. If setted the content of ck is relative.
     * @param aRelk 0 not present in the ctin-m directive, -1 absolute displacement, 1 relative displacement.
     */
    void setRelk(short aRelk){ relk=aRelk; }
    
    /** 
     * Get relative or absolute displacement for k coordinate. 
     * @return 0 not present in the ctin-m directive, -1 absolute displacement, 1 relative displacement.
     */
    short getRelk(){ return relk; }
    
    /** 
     * Set relative or absolute displacement for t or T coordinate. If setted the content of ct is relative.
     * @param aRelt the relt value (look to the definition of the attribute).
     */
    void setRelt(short aRelt){ relt=aRelt; }
    
    /** 
     * Get relative or absolute displacement for t or T coordinate. 
     * @return the value of relt (look to the definition of the attribute).
     */
    short getRelt(){ return relt; }
    
    /** 
     * Set the type of ctin-m. If setted the connection present in the description file is ctin otherwise is ctinm.
     * @param aType the type of ctin-m.
     */
    void setCtinType(bool aType){ ctinType=aType; }
    
    /** 
     * Get the type of ctin-m. 
     * @return the type of ctin-m: true ctin and false ctinm.
     */
    bool isCtinType(){ return ctinType; }
    
    /**
     * We set a RELATIVE interval that is only allowed for the coordinates (i,j,k,t), so we set the range of the coordinates. 
     * We call this function only if in the description file the connection ctin-m is relative.
     * @param rangeOp is the range operator that is one of the choice of possible interval: "", "+", "-", "~", "%+", "%-", "%~", "#+", "#-", "#~".
     * @param value is the value parsed in the description file inside the directive ctin.
     * @param parameterNumber is a number that identifies the actual position we are parsing in the directive. 
     * This is why we need to understand which kind of range we are managing: 
     * 0 is for the input range, 1 is for the output range, 2/3/4/5/6 are for the i/j/k/t/T respectively.
     */
    void setRelativeInterval(string rangeOp, int value, short parameterNumber);
     
    /**
     * We set an ABSOLUTE interval, so we set the range of the coordinates and of the input/output. 
     * We call this function only if in the description file the connection ctin-m is absolute (i.e.: 2..4)
     * @param aBegin is the begin value of the range (in the example is 2).
     * @param anEnd is the end value of the range (in the example is 4).
     * @param parameterNumber is a number that identifies the actual position we are parsing in the directive. 
     * This is why we need to understand which kind of range we are managing: 
     * 0 is for the input range, 1 is for the output range, 2/3/4/5/6 are for the i/j/k/t/T respectively.
     */
    void setAbsoluteInterval(int aBegin, int anEnd, short parameterNumber);

    /**
     *  SOME CHECK for the option specified on the directive ctin-m. Here we check some consistency parameters of the ctin-m.
     *  The consistency of the ctin-m syntax is already tested by the grammar. Here we test the semantic of the parameters.
     *  The only thing that we do not test here is the consitency between directive ctin-m. Each test concern one ctin-m, 
     *  the modules and the relative space and trajetory.
     *  @param inMod is a pointer on the object input module.
     *  @param outMod is a pointer on the object output module.
     *  @param theTableSpace is the object that contain all the spaces.
     *  @param theOperatorTable is the object that contain all the operators.
     *	@param theTrajectoryTable is the object that contain all the Trajectories.
     */
    void checkOptions(Modul *inMod, Modul *outMod, Table<Space> theSpaceTable, Table<Operator> theOperatorTable, Table<Trajectory> theTrajectoryTable);

    /**
     * Allows to have a nice and automatic visualization of the Connection object elements.
     * @return the number of significant properties stored in this class.
     */
    static int getPropertyCount() { return 8; }

    /**
     * Allows to have a nice and automatic visualization of the Connection object elements. 
     * In this method we retrive this kind of propertis: "from", "N� output", "Ci", "Cj", "Ck", "Ct", "to", "N� input".
     * These are prefixed string that never change.
     * @param anIndex is the number of property required.
     * @return a string that represent the number of property stored in this class.
     */
    static std::string getPropertyName(int anIndex){
      ENFORCE(anIndex >= 0 && anIndex < Connection::getPropertyCount());
      static const char* propertyName[] = { "from ", "N output", "Ci", "Cj", "Ck", "Ct", "to ", "N input" };
      return propertyName[anIndex];
    }

    /**
     * Allows to have a nice and automatic visualization of the Connection object elements. 
     * In this method we retrive the value of significant property: "from", "N� output", "Ci", "Cj", "Ck", "Ct", "to", "N� input".
     * In function of the integer in input to the method we retrieve the respective property.
     * @param anIndex is the number of property required.
     * @return the property stored in this class.
     */
    std::string getProperty(int anIndex) const;

  private:
    
    /**
     * Set the value in the right vector in function of the coordinate number "parameterNumber". 
     * In this method "parameterNumber" can be addressed only with coordinate i,j,k,t or T and not with input or output modules.
     * @param value the coordinate value to be setted.
     * @param parameterNumber is a number that identifies the actual position we are parsing in the directive. 
     * This is why we need to understand which kind of range we are managing: 
     * 0 is for the input range, 1 is for the output range, 2/3/4/5/6 are for the i/j/k/t/T respectively.
     */
    void setCoordinateValue(int value, short parameterNumber);
    
    /**
     * Allows to test the translation, between two trajectories, of one instant of time.
     * @param theTrajectoryTable is the object that contain all the Trajectories.
     * @param outTraj is the trajectory of the output module.
     * @param inInstant is the instant to be translated. This instant belong to the input module trajectory.
     * @param inTraj is the trajectory of the input module.
     * @return the equivalent instant of the output trajectory.
     */
    int Ypdt2pdt(Table<Trajectory> theTrajectoryTable, string outTraj, int inInstant, string inTraj);

    string inModule, outModule;			      //!< Names of the modules connected.
    int inBegin, outBegin;			      //!< Begin number of input and output range connection.
    int inEnd, outEnd;				      //!< End number of input and output range connection. If 0 there is no interval.
    
    //! Relative or absolute displacement for i,j,k coordinates.
    //! 0 not present in the ctin-m directive, -1 absolute displacement, 1 relative displacement.
    short reli, relj, relk;			      
    
    //! Relative or absolute displacement for t and T coordinates.
    //! 0  ==> t not present in the ctin-m directive;
    //! -1 ==> absolute displacement;
    //!  1 ==> relative displacement;
    //!  2 ==> relative displacement if the input trajectory is different from the output trajectory and we have 't' in the description file.
    //! In this case we have to translate the instants from the inTraj to the outTraj.
    short relt;
    bool ctinType;  //!< True if ctin, false if ctinm.

    bool t; //!< True if t has been inserted in the description file.
    bool T; //!< True if T has been inserted in the description file.
  };

/**
 * Allows to have a nice and automatic visualization of the Connection object elements. 
 * In this method we retrive the value of significant property: "from", "N� output", "Ci", "Cj", "Ck", "Ct", "to", "N� input".
 * In function of the integer in input to the method we retrieve the respective property.
 * @param anIndex is the number of property required.
 * @return the property stored in this class.
 */
inline
std::string Connection::getProperty(int anIndex) const{
  std::string property;
  switch (anIndex){
    case 0:
      property = outModule;
      break;
    case 1:
      property = stream_cast<std::string>(outBegin);
      break;
    case 2:
      if(reli==1){
	if(ci[0]>=0) // Just for output indentation.
	  property=" r  " + stream_cast<std::string>(ci[0]);
	else property=" r " + stream_cast<std::string>(ci[0]);
      }else if(reli==-1){
	if(ci[0]>=0) // Just for output indentation.
	  property=" a  " + stream_cast<std::string>(ci[0]);
	else property=" a " + stream_cast<std::string>(ci[0]);
      }else if(reli==0) property="0"; // Just for homogeneity, is always true.
      break;
    case 3:
      if(relj==1){ 
	if(cj[0]>=0)
	  property=" r  " + stream_cast<std::string>(cj[0]);
	else property=" r " + stream_cast<std::string>(cj[0]);
      }else if(relj==-1){ 
	if(cj[0]>=0)
	  property=" a  " + stream_cast<std::string>(cj[0]);
	else property=" a " + stream_cast<std::string>(cj[0]);
      }else if(relj==0) property="0";
      break;
    case 4:
      if(relk==1){
	if(ck[0]>=0)
	  property=" r  " + stream_cast<std::string>(ck[0]);
	else property=" r " + stream_cast<std::string>(ck[0]);
      }else if(relk==-1){
	if(ck[0]>=0)
	  property=" a  " + stream_cast<std::string>(ck[0]);
	else property=" a " + stream_cast<std::string>(ck[0]);      
      }else if(relk==0) property="0";
      break;
    case 5:
      if(relt==1){ 
	if(!ct[0])
	  property=" r  " + stream_cast<std::string>(ct[0]);
	else property=" r " + stream_cast<std::string>(ct[0]);
      }else if(relt==-1){
	if(!ct[0])
	  property=" a  " + stream_cast<std::string>(ct[0]);
	else property=" a " + stream_cast<std::string>(ct[0]);
      }else if(relt==2){
	if(!ct[0])
	  property=" R  " + stream_cast<std::string>(ct[0]);
	else property=" R " + stream_cast<std::string>(ct[0]);
      }else if(relt==0) property="0";
    break;
    case 6:
      property = " " + inModule;
      break;
    case 7:
      property = stream_cast<std::string>(inBegin);
      break;
  }

  return property;
}

} // End namespace yao.
#endif // !YAO_INCLUDE_HPP_INCLUDED.
