//! \file    Operator.hpp
//! \brief   Declaration and implementation of the (linear) Operator class,
//!          which handles memory space and computation.
//! \version 2007/10/01 (yyyy/mm/dd) 
//! \author  Luigi Nardi <luiginardi(at)gmail.com> 
//
// This file is part of YAO, a framework for variational assimilation in
// numerical models.
// Copyright (c) 2001-onwards LOCEAN. All rights reserved.
//
// This program may be redistributed and/or modified under the terms of the GNU
// General Public License as published by the Free Software Foundation, either
// version 2 of the license, or (at your option) any later version.
//
// This program is provided AS IS and WITHOUT ANY WARRANTY, including any
// implied warranty of DESIGN, MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.


#ifndef YAO_OPERATOR_HPP_INCLUDED
#define YAO_OPERATOR_HPP_INCLUDED

#include "Component.hpp"
#include "SpaceExtent.hpp"
#include "Trajectory.hpp"

#include <iterator>
#include <sstream>

namespace yao{

  //! (Linear) Operator class, handles memory space and computation.
  class Operator: public Component<Trajectory>, public SpaceExtent{

    public:

      //! Supported types.
      enum Type{
	M = 'M', //!< Model operator (default).
	H = 'H', //!< Observation operator.
	B = 'B', //!< Background error covariance operator.
	R = 'R', //!< Observation error covariance operator.
	K = 'K'  //!< Balance operator.
      };

      /**
       * Default constructor.
       * @param aName is the name of the Operator.
       */
      Operator(const std::string& aName=""): 
	Component<Trajectory>(aName, 0), counterOrderHeader(false), counterOrderBody(false), theType(Operator::M) {}

      /**
       * Constructor.
       * @param aName is the name of the Operator.
       * @param aParent is the Trajectory linked to the Operator.
       * @param aShape is the vector that contains the dimension of the Operator on all axes.
       * @param aType is the operator type (see YAO documentation).
       */
      Operator(const std::string& aName, const Trajectory* aParent, std::vector<int> aShape, Type aType=Operator::M):
	Component<Trajectory>(aName, aParent), SpaceExtent(aShape), counterOrderHeader(false), counterOrderBody(false), theType(aType){}

      /**
       * Set the space counter of the directive order.
       * true the space has been already inserted in the header of the directive order modinspace, false otherwise.
       * With "in the header" we mean the first line of the directive order: order modinspace space_name....
       * This is in opposite with the body of the directive that is the rest of the directive.
       * With this header example space_name will be marked to true.
       * We can find spaces in the header modinspace directive and in the body of spaceintraj directive. 
       * This counter allows to perform the control on the header of modinspace.
       * @param flag is the value to be setted.
       */
      void setCounterOrderHeader(bool flag){ counterOrderHeader = flag; }

      /**
       * Check if the space has been already inserted in an header of the directive modinspace order.
       * @return true the object has been already inserted in the directive, false otherwise.
       */
      bool isCounterOrderHeader(){ return counterOrderHeader; }

      /**
       * Set the space counter of the directive order.
       * true the space has been already inserted in the body of the directive order spaceintraj, false otherwise.
       * With "in the body" we mean all that is not the header (look setCounterHeader).
       * We can find spaces in the header modinspace directive and in the body of spaceintraj directive. 
       * This counter allows to perform the control on the body of spaceintraj.
       * @param flag is the value to be setted.
       */
      void setCounterOrderBody(bool flag){ counterOrderBody = flag; }

      /**
       * Check if the space has been already inserted in a body of the directive trajinspace order.
       * @return true the object has been already inserted in the directive, false otherwise.
       */
      bool isCounterOrderBody(){ return counterOrderBody; }

      /**
       * Get the type of the Operator.
       * @return the type.
       */
      Type getType() const { return theType; }

      //! The number of information stored in this class.
      static int getPropertyCount() { return 4; }

      //! Get the name of the property: "operator/space", "shape", "parent", "type" 
      static std::string getPropertyName(int anIndex){
	ENFORCE(anIndex >= 0 && anIndex < Operator::getPropertyCount());
	static const char* propertyName[] = { "operator/space", "shape", "parent", "type" };
	return propertyName[anIndex];
      }

      //!Get the property information in function of an integer that identifies it.
      std::string getProperty(int anIndex) const;

    private:
      bool counterOrderHeader;  //!< A counter to control that the directive order modinspace include no more than one time this object.
      bool counterOrderBody;    //!< A counter to control that the directive order spaceintraj include no more than one time this object.
      Type theType; //!< Type.
  };

  inline std::string Operator::getProperty(int anIndex) const{
    std::string property = Object::getProperty(anIndex);
    std::ostringstream interpreter;
    switch (anIndex){
      case 1:
	std::copy(theShape.begin(), theShape.end(), std::ostream_iterator<int>(interpreter, " "));
	property = interpreter.str();
	property.erase(property.length() - 1);
	break;
      case 2:
	property = stream_cast<std::string>(theParent->getName());
	break;
      case 3:
	property = stream_cast<std::string>(static_cast<char>(theType));
    }

    return property;
  }

  typedef Operator Space; //!< So-called space (model operator);

} // End namespace yao.
#endif //! YAO_OPERATOR_HPP_INCLUDED.
