//! \file    Order.hpp
//! \brief   Declaration and implementation of the Order class, which
//!          handles order of computation between modules, spaces and trajectories.
//! \version 2008/07/24 (yyyy/mm/dd) 
//! \author  Luigi Nardi <luiginardi(at)gmail.com> 
//
// This file is part of YAO, a framework for variational assimilation in
// numerical models.
// Copyright (c) 2001-onwards LOCEAN. All rights reserved.
//
// This program may be redistributed and/or modified under the terms of the GNU
// General Public License as published by the Free Software Foundation, either
// version 2 of the license, or (at your option) any later version.
//
// This program is provided AS IS and WITHOUT ANY WARRANTY, including any
// implied warranty of DESIGN, MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.

#ifndef YAO_ORDER_HPP_INCLUDED
#define YAO_ORDER_HPP_INCLUDED

#include <cstdio>
#include <cstring>
#include <cstdlib>
#include "../help/Helper.hpp"
#include <iostream>
#include <string>
#include <vector>

#include "Table.hpp"
#include "Modul.hpp"
#include "Connection.hpp"
#include "Operator.hpp"

// ----------------- AJOUT CB HEADER --------------------//

#include "Connection.hpp"
#include "Operator.hpp"
#include <fstream>

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/strong_components.hpp>
#include <boost/graph/topological_sort.hpp>

namespace boost {

        /// A property that allow to record in each vertex, after a topological sort was performed, its discover time.
        enum vertex_topo_order_t        { vertex_topo_order      = 113 };
        /// A property that allow to record in each vertex, aftter a component seek was performed, the component where it is located. 
        enum vertex_affected_comp_t     { vertex_affected_comp   = 114 };
        /// A property that allow to record in each vertex its original Id as it appear in the original graph.
        enum vertex_original_id_t       { vertex_original_id =  115 };
        /// A property that allow to record in each edge the sign if the first dimension (i) as it come from the ctin statements.
        enum edge_weight_d1_t           { edge_weight_d1   = 117 };
        /// A property that allow to record in each edge the sign if the second dimension (j) as it come from the ctin statements.
        enum edge_weight_d2_t           { edge_weight_d2   = 118 };
        /// A property that allow to record in each edge the sign if the third dimension (k) as it come from the ctin statements.
        enum edge_weight_d3_t           { edge_weight_d3   = 119 };

        BOOST_INSTALL_PROPERTY(vertex, topo_order);
        BOOST_INSTALL_PROPERTY(vertex, affected_comp);
        BOOST_INSTALL_PROPERTY(vertex, original_id);

        BOOST_INSTALL_PROPERTY(edge, weight_d1);
        BOOST_INSTALL_PROPERTY(edge, weight_d2);
        BOOST_INSTALL_PROPERTY(edge, weight_d3);

        }

// --------------- FIN AJOUT CB HEADER ------------//


#define NB_MAX_ORD_TOK  800  //<! Max number of token for the order directive.

using namespace std;

/// The struct that follows allows to reverse the vector of tokens (forward <-> backward)
struct st_token {
       char token[81];   
       int level;
       struct st_token *brother;
       struct st_token *father;
       struct st_token *son;
     };

namespace yao{

  /** 
   *  Order class, handles the directive order.
   *  
   *  It allows to stock all the information about the directive order of the description file.
   *  It allows also to perform the basic operations get and set of the members and some other operations on these members like
   *  test on the options of the order directive.
   *
   *  For each directive order we have a new Order object that manage it.
   */
  class Order {
  
  public:

    /**
     * Is the vector that contains all the tokens of the order. 
     * i.e.: we have an order like this: 
     * order modinspace space
     *	order YA1 YA2
     *	  A B C
     *	forder
     * forder.
     * The vector will contain: order, YA1, YA2, A, B, C, forder. Each element is called token.
     */
    vector<string> orderTokens;

    //! Default constructor.
    Order(Table<Modul> * aTheModulTable){ theModulTable = aTheModulTable; }

	Order(){ }

    
/*    //! Constructor for the order coherences check
    Order(){ }
   
    Order(Table<Connection> * aConnectionTable, Table<Modul> * aModulTable, Table<Space> * aSpaceTable, 
	Table<Operator> * aOperatorTable) : theConnectionTable(aConnectionTable), theModulTable(aModulTable),
        theSpaceTable(aSpaceTable), theOperatorTable(aOperatorTable) { }
*/


    /** 
     *	Set the name of the Space or Trajectory depending on orderPhase variable. 
     *	The name of the order is the type of order we are performing so modinspace or spaceintraj
     *	@param aName the name.
     */
    void setName(string aName){ name = aName; } 

    /** 
     *	Set the value relative to the type of phase (that is also the type of order defined by the user in almost all cases). 
     *	The phases are: 1 modinspace, 2 spaceintraj, 3 end of all orders (a final block).
     *	@param value the phase value.
     */
    void setOrderPhase(int value){ orderPhase = value; } 

    /** 
     *	Get the name of the Space or Trajectory depending on orderPhase variable.
     *	@return the name of the Space or Trajectory.
     */
    string getName(){ return name; } 
    
    /** 
     *	Get the integer relative to the type of phase we are attemping on (is also the type of order defined by the user in almost all cases). 
     *	The phases are: 1 modinspace, 2 spaceintraj, 3 end of all orders (a final block).
     *	@return the phase value.
     */
    int getOrderPhase(){ return orderPhase; } 

    /** 
     *	Translate the value in input in a string and then returns it.
     *	The integer value in input is one of these possibles cases:
     *	YA1 = 1 ; YB1 = -1 ; YA2 = 2 ; YB2 = -2 ; YA3 = 3 ; YB3 = -3 ;
     *	@param anAxe the integer token to be firstly translated.
     *	@return the string that represents the axe.
     */
    static string translateAxe(int anAxe);
    
    //-----------------------------------------------------------------------------------
    // The part that follow allow to reverse the vector of tokens (forward <-> backward)
    
    /**
     * Put in the opposite order the vector of tokens of a directive order modinspace or spaceintraj.
     * i.e.: we have an order like this: 
     * order modinspace space
     *	order YA1 YA2
     *	  A B C
     *	forder
     * forder.
     * The vector will contain: order, YA1, YA2, A, B, C, forder.
     * The opposite vector will contain: order, YB1, YB2, C, B, A, forder. 
     * This is done in order to generate the code for the backward function. We could reverse a lot of time the vector of tokens and 
     * each time we obtain the precedent configuration.
     */
    void backOrder();

     
// ------------------  AJOUT CB VERIFICATION DES BOUCLES ------------------ //

  public : 

  void checkOrder(Table<Connection>& theConnectionTable, Table<Modul>& theModulTable, Table<Space>& theSpaceTable, Table<Operator>& theOperatorTable, Table<Order>& theOrderTable);

  private : 

  int getDimSize(string space, int dim, Table<Space>&  theSpaceTable);

  int checkConnection(stringstream& theConnection, stringstream& errorMsg, stringstream& warningMsg, int numDim, Table<Modul>::iterator itSource, Table<Modul>::iterator itTarget, int valCijk[3], bool absCijk[3], int lastAxis, Table<Space>&  theSpaceTable); 

  int checkModOrder(stringstream& theConnection, stringstream& errorMsg, Table<Modul>::iterator itSource, Table<Modul>::iterator itTarget);

  Table<Modul>::iterator getModule(string nom, Table<Modul>& theModulTable);    // Return an iterator on the module named "nom".

  /**
   *  Set up all the loops by scanning the order directives.
   *  For each module, adds order information (axis order and direction, sequence with other modules, loop identifier )
   */
  void setLoop(Table<Connection>& theConnectionTable, Table<Modul>& theModulTable, Table<Space>& theSpaceTable, Table<Operator>& theOperatorTable, Table<Order>& theOrderTable);

  /**
   * Check all connections defined in the ctin directives.
   */
  bool checkConnections(stringstream& errorMsg, stringstream& warningMsg, Table<Connection>& theConnectionTable, Table<Modul>& theModulTable, Table<Space>&  theSpaceTable, Table<Order>& theOrderTable);


// ----------------- FIN AJOUT CB VERIFICATION DES BOUCLES ----------------//


// -------------- AJOUT CB GENERATION AUTOMATIQUE DE L'ORDER --------------//


   // vector<string> myOrderTokens;

   /**
    *  dedicated light structures that register, from the yaoo objects,  all the informations needed to perform the order generation
    */
   typedef pair<int, int> edge;

   /** 
    * Definition of the vertices properties embedded in the graph type that will be used in all the order generation functions.
    */  
   typedef boost::property<boost::vertex_name_t, string,
                           boost::property<boost::vertex_topo_order_t, int,
                           boost::property<boost::vertex_affected_comp_t, int,
                           boost::property<boost::vertex_original_id_t, int > > > > verticesProperties;

   /**
    *  Definition of the edges properties embedded in the graph type that will be used in all the order generation functions.
    */    
   typedef boost::property<boost::edge_weight_d1_t, int,
                           boost::property<boost::edge_weight_d2_t, int,
                           boost::property<boost::edge_weight_d3_t, int > > > edgesPorperties;

   /**
    * Definition of the graph type that will be used in all the order generation functions.
    */ 
   typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::bidirectionalS, verticesProperties, edgesPorperties> myGraphType;

   /**
    *  Attribute type of the cfc structures that embed the axis number as first element and the axis value as second element.
    */
   typedef pair<int, int> valAxisType;

   /**
    *    Structure that record infromations about one component
    */ 
   struct cfcStruct { vector<valAxisType> cfcSign; vector<int> embeddedModuls; };

   /**
    *  Definition of the data type that record all the component invormations of a graph (vector of cfcStruct).
    */
   typedef vector<cfcStruct> cfcStructTableType;

   
   typedef cfcStructTableType** middleCompType;

   typedef cfcStructTableType**** innerCompType;

   /**
    *  Graph iterators and descriptors that are used all along the program (declared here to avoid several decalarations inside the source code).
    */ 

   typedef boost::graph_traits<myGraphType>::edge_descriptor edgeDescriptor;

   typedef boost::graph_traits<myGraphType>::vertex_descriptor vertexDescriptor;

   boost::graph_traits<myGraphType>::vertex_iterator vertexIterOne, vertexIterTwo;

   boost::graph_traits<myGraphType>::edge_iterator edgeIterOne, edgeIterTwo;

   /**
    *  A string that contain the project name given as a parameter to automaticOrderGenerator (declared here to avoid changing all the function calls)
    */
   string theProject;

    public:

        /**
         * Get informations on connections, modules and spaces then launch all the procedures used to generate the order statments. 
         * The relevant informations are extracted from the existing tables then integrated in dedicated light structures.
         * The dedicated structures are declared as global datas then accessed and updated by the functions. 
         */
         void automaticOrderGenerator(string projectName, Table<Order>& theOrderTable, Table<Modul>& myModulTable, Table<Connection>& myConnectionTable, Table<Space>& mySpaceTable);

    private:

        // -- Functions that build and populate the data structures --

        /**
        * Function that manage the 3D components by performing the following tasks : 
        *       - get all the components from the main graph (tarjan component search)
        *       - reset the outerComp vector size accordignly to the number of components detected
        *       - for each component, get the sign of the three axis 
        *       - for each component, sort the axis accordignly to their signs 
        *       - for each component, detect the errors (all errors can be detected here, no need to perform more controls after)
        *       - reduce the main graph by keeping just one vertex by component 
        *       - perform a topological sort on this reduced graph to get a component ordonnancing 
        *       - sort the outerComp vector elements accordignly to the component ordonnancing 
        *       - for each component, subset the main graph for getting a graph with just this component (3D component graph)
        *       - for each component, launch the function that will manage the two dimensionals components with the 3D component graph as a parameter
        */
        void setOuter(cfcStructTableType& outerComp, middleCompType& middleComp, innerCompType& innerComp, myGraphType& myGraph);

        /**
        * Function that manage the 2D components by performing the following tasks : 
        *       - get a 3D component graph from setOuter(), this one is given as a parameter
        *       - fix the dimension "fixedDim" given as a parameter (cf fixDim) to obtain a 2D fixed graph (from the 3D component graph)
        *       - get all the components from the 2D fixed graph (tarjan component search)
        *       - reset the middleComp[noComp3D][fixedDim] vector size accordignly to the number of components detected
        *       - for each component, get the sign of the two axis 
        *       - for each component, sort the axis accordignly to their signs 
        *       - reduce the 2D fixed graph by keeping just one vertex by component 
        *       - perform a topological sort on this reduced graph to get a component ordonnancing 
        *       - sort the middleComp[noComp3D][fixedDim] vector elements accordignly to the component ordonnancing 
        *       - for each component, subset the main graph for getting a graph with just this component (2D component graph)
        *       - for each component, launch the function that will manage the one dimensional components with the 2D component graph as a parameter
        * @param noComp an integer value that index the 3D component where the current 2D component vector come from
        * @param fixedDim an integer value that index the first dimension that have been used in the 3D component to get the 2D component
        * @param dimOne an integer value that indicate which is the second value that must be fixed in priority (could be found also from fixedDim) 
        * @param dimTwo an integer value that indicate which is the third value that must be fixed in last (could be found also from fixedDim and dimOne)
        * @param currentGraph the 3D component graph from which the 2D component vector will be extracted
        */
        void setMiddle(int fixedDim, int dimOne, int dimTwo, myGraphType& currentGraph, int noComp, middleCompType& middleComp, innerCompType& innerComp);

        /**     
        * Function that manage the 1D components by performing the following tasks : 
        *       - get a 2D component graph from setMiddle()
        *       - fix the dimension "noFixed2" from the 2D component graph to obtain a 1D fixed graph
        *       - get all the components from the 1D fixed graph (tarjan component search)
        *       - reset the innerComp[noComp3D][noFixed1][noComp2D][noFixed2] vector size accordignly to the number of components detected
        *       - for each component, get the sign of the axis
        *       - keep only edges with null values on the 1D fixed graph then perform a topological sort on this graph
        *       - for each component, use this topological sort for ordonnancing the moduls inside the component (sort the embeddedModuls vector)
        *       - reduce the 1D fixed graph by keeping just one vertex by component 
        *       - perform a topological sort on this reduced graph to get a component ordonnancing 
        *       - sort the innerComp[noComp3D][noFixed1][noComp2D][noFixed2] vector elements accordignly to the component ordonnancing 
        * @param noComp3D an integer value that index the 3D component where the current 1D component vector come from
        * @param noFixed1 an integer value that index the first dimension that have been used in the 3D component to get the 2D component
        * @param noComp2D an integer value that index the 2D component where the current 1D components vector come from
        * @param noFixed2 an integer value that index the second dimension n the 2D component to get the current 1D component vector
        * @param freeDim an integer value that indicate which is the actual dimension for the 1D components vector (could be found also from noFixed1 and noFixed2) 
        * @param upperGraph the 2D component graph from which the 1D component vector will be extracted
        */
        void setInner(int noComp3D, int noFixed1, int noComp2D, int noFixed2, int freeDim, myGraphType& upperGraph, innerCompType& innerComp);

        // --  Functions for managing the graphs --

        /**
        * Function for setting up the main graph.
        * This function create a graph then populate its properties accordignly to the boost API.
        * @param currentGraph a reference on the graph that should be created.
        * @param valCi a vector of integer that register the connection values for the i dimension.
        * @param valCj a vector of integer that register the connection values for the j dimension.
        * @param valCk a vector of integer that register the connection values for the k dimension.
        * @param modulList a vector of strings that regster the modul names.
        * @param edgeArray a vector of edges that register the source modul and the target modul of each edge (connection). 
        */
        void generateGraph(myGraphType& currentGraph, const vector<int>& valCi, const vector<int>& valCj, const vector<int>& valCk, const vector<string>& modulList, const vector<edge>& edgeArray);

        /**
         * Function that make a deep copy from the graph upperGraph to the graph outGraph
         * The function is void and the graph outGraph have to be given as a parameter because there is some problems in getting a graph from a function  
         * This function perform a deep copy of the graph : edges and vertices but also all properties are copied
         * @param upperGraph a reference to the graph that must be copied (not declared const as it should be because boost seems to forbid it)
         * @param outGraph a reference to the graph that must receive all its informations from the copied graph 
         */
        void copyGraph(myGraphType& upperGraph, myGraphType& outGraph);

        /**
        * Function that keep only the edges with zero value for the dimension given as a parameter 
        * As a result of this fixation, whe know that remaining connected vertices (moduls in YAO) are linked on the same node for the fixed dimension 
        * @param currentGraph a reference to the graph that must have one of its dimension fixed
        * @param noFixedDim an integer value indicating which dimension have to be fixed
        */
        void fixGraph(myGraphType& currentGraph, int noFixedDim);

        /**
        * Function that create a graph which vertices are the components of another graph
        * @param nbComp the number of components included in the innerGraph
        * @param innerGraph the graph whose components have to be reduced as single vertices 
        * @param outerGraph the reduced graph, which vertices are the components coming from the graph innerGraph 
        */
        void reduceGraph(myGraphType innerGraph, myGraphType& outerGraph, int nbComp);

        /**
        * Function that is used to extract a component from a graph 
        * This function create a subset graph that just include vertices (and edges) from the "noComp" component 
        * @param noComp the component that must be extracted from graph "upperGraph" 
        * @upperGraph a reference to the graph from which the component have to be extracted 
        */
        void keepComponent(int noComp, myGraphType& upperGraph);

        /**
        * Function that reset the main graph and all its properties. 
        * This function make a tabula rasa of the graph that have been created for the latest space
        * and allow it to register informations on the space that is currently analyzed.
        * As informations on the graph are just used to create the current modinspace statement, there 
        * is no need to keep those informations from space to space. 
        * As as the main graph (myGraph) is a global data, it is very important to perform  this operation.
        * @currentGraph a reference to the graph that must be resetted
        */
        void resetGraph(myGraphType& currentGraph);


        // -- Functions for managing axis (sign and ordering) --

        /**
        * Function that perform an offset on the the data structures Ci, Cj and Ck before to create the main graph. 
        * This offset tend to treat all the main graphs as 3D ones (and then to use the same entry function for both). 
        * If the graph has just one dimension, then this dimension is moved from Ci to Ck then Ci and Cj are padded with zero.
        * If the graph has just two dimensions, then Ci and Cj are moved to Cj and Ck then Ci is padded with zero.
        * As a result, Ci, Cj and Ck will allow to create a 3D graph with just one main 3D component if the original graph 
        * is a 2D graph and with just one  main 3D component and one main 2D component if the original graph is a 1D graph. 
        * As there is just one component before the releving graph, there is no overhead. The only adaptation will be for the 
        * reading  function that will be dedicated to one, two or three dimensions.
        * @param nbDim an integer value that indicate the number of dimensions on which is defined the original graph
        * @param valCi a reference to the vector that register values of connections for the dimension i
        * @param valCj a reference to the vector that register values of connections for the dimension j
        * @param valCk a reference to the vector that register values of connections for the dimension k
        * @param nbEdges an integer value that indicate the number of edges (connections) of the graph
        */
        void to3D(int nbDim, vector<int>& valCi, vector<int>& valCj, vector<int>& valCk, int nbEdges);

        /**
        * Function that fix a dimension of a graph then perform a sign analysis after fixation 
        * The fixation of the dimension "fixedDim" from the setOuter() function act as a forward checking on 2D components 
        * to allow to register informations on the second axis (dimension and direction) 
        * @param fixedDim an integer value indicating which dimension must be fixed
        * @param freeDim an integer value indicating on which dimension the sign must be analyzed 
        * @param noComp an integer value indicating on which component the fixation and the sign analysis must be performed
        * @param upperGraph a reference to the graph on which the function must work 
        */
        int getImposed(int fixedDim, int freeDim, int noComp, myGraphType& upperGraph);

        /**
        * Function that fix two dimensions of a graph then perform a sign analysis after fixations
        * The fixation of the dimensions "fixedDimOne" and "fixedDimTwo" from the setOuter() function act as a forward checking on 1D components
        * to allow to register informations on the third axis (dimension and direction) 
        * @param freeDim an integer value indicating which dimension is the last one (so it allows to know the two others must be fixed)
        * @param noComp an integer value indicating on which component the fixations and the sign analysis must be performed
        * @param upperGraph a reference to the graph on which the function must work 
        */
        int getImposed(int freeDim, int noComp, myGraphType& upperGraph);

        /**
        * Function used to know if the third axis have to be decomposed 
        * This function reduce two times the main graph then seek a common value on the remaining edges 
        * If there is no common value the axis have to be decomposed 
        * If there is a common value or if there is no remaining edges in the last fixed graph, then the axis does not have to be decomposed
        * @param freeDim an interger value indicating on which dimension the sign must be analyzed
        * @param upperGraph a reference to the graph on which the operations have to be performed
        */
        bool isToDecompose(int freeDim, myGraphType& upperGraph);

        bool isToDecompose2D(int fixedDim, int noComponent,  myGraphType& upperGraph);

        /**
        * Function that get the axis sign for the dimension "noDim" on the edes of the component "noComp" in the graph "currentGraph"
        * return 0 if there is no edges, -1 or 1 if there is a common sign and -2 if there is no common sign 
        * @param noDim an interger value indicating the dimension on which the sign must be analyzed
        * @param noComp an interger value indicating the component of the graph that must be analyzed
        * @param currentGraph a reference to the graph containing the component that must be analyzed
        */
        int getCommonSign(int noDim, int noComp, myGraphType& currentGraph);

        /**
        * Function that get the axis sign for the dimension "noDim"  in the graph "currentGraph"
        * return 0 if there is no edges, -1 or 1 if there is a common sign and -2 if there is no common sign 
        * This function bring the same information than getCommonSign() with three parameters except it does not seek a special component
        * @param noAxis an interger value indicating the dimension on which the sign must be analyzed
        * @param cfixedGraph a reference to the graph on which the sign must be analyzed
        */
        int getCommonSign(int noAxis,  myGraphType& fixedGraph);

        /**
        * Function that get the axis sign for the dimension "noFree" in the graph "currentGraph"
        * return 0 if there is no edges, -1 or 1 if there is a common sign and -2 if there is no common sign 
        * This function bring the same informations as "getCommonSign()" but is dedicated to the "innerComp" array because 
        * she fix the dimension "noFixed" of the 2D component graphs given as parameter then perform the sign analyzis on the "noFree" dimension 
        * @param noFixed1 an integer value indicating which dimension have to be fixed on the graph 
        * @param noFree and integer value indicating on which dimension the sign must be analyzed
        * @param currentGraph a reference to the graph on which the fixation and the sign search have to be performed
        */
        int getCommonSignInner(int noFixed, int noFree, myGraphType& currentGraph);

       /**
        * This function perform almost the same operations than getCommonSign() except the target modul does not have to be in 
        * the checked component. 
        */
        int getCommonFixedSign(int noDim, int noComp, myGraphType& currentGraph);


        // -- Functions that create the hidden file and populate the order Table by reading the data structures -- 

       /*
        * Function that read one order statement and his parameter ("order Y**") 
        * @param noAxis the axis dimension number on which the order have to be displayed (0 = i, 1 = j, 2 = k)
        * @param valAxis the axis value on which the order have to be displayed (used to display A or B)
        * @param noParam the parameter number on which the order have to be displayed (is the parameter is the first, second or third)           
        */
        void readParam(int noAxis, int valAxis, int noParam);

        /**
        * Function that perform the reading of all the order statement the same way that in the read() function 
        * The difference is that, as the first and second axis are dummies, just one component of the 3D components vector 
        * and one component of the 2D components vector (for the first axis as fixed dimension) are considered.
        * The true reading begins from the 1D components vector the same way than for the read() funtion except 
        * for the axis number that have an offset because axis "k" is in reality "i" (cf. to3D() function). 
        */
        void read1D(cfcStructTableType outerComp, middleCompType& middleComp, innerCompType& innerComp, myGraphType& myGraph);

        /**
        * Function that perform the reading of all the order statement the same way that in the read() function 
        * The difference is that, as the first axis is dummy, just one component of the 3D components vector is considered.
        * The true reading begins from the 2D components vector the same way than for the read() funtion except 
        * for the axis number that have an offset because "j" and "k" axis are in reality "i" and "j" axis (cf. to3D() function). 
        */
        void read2D(cfcStructTableType outerComp, middleCompType& middleComp, innerCompType& innerComp, myGraphType& myGraph);

        /** 
        * Function that read the three structures ("outerComp", "middleComp" and "innerComp") then display all the order 
        * statement accordignly to the values read in the "cfcSign" attributes of the components structures 
        * If all informations about the order (axis ordonnancing and path) can be found in the 3D component vector, read 
        * them from here then cross "middleComp" and "innerComp" just for displaying the corresponding moduls. 
        * If informations on the second and third axis can not be found in the 3D component vector, read them from the 2D 
        * components vector (middleComp) then cross "innerComp" just for displaying the corresponding moduls. 
        * If informations on the third axis can not be found in the 3D component vector or in the 2D component vector, 
        * read them and the moduls from the 1D component vector (innerComp).
        */
        void read3D(cfcStructTableType outerComp, middleCompType& middleComp, innerCompType& innerComp, myGraphType& myGraph);

        // -- Functions that display data structures -- 

        /**
        * Function that show vertices and edges of a graph and their attributes
        * @param currentGraph a reference to the graph on which informations have to be shown
        */
        void showGraph(myGraphType& currentGraph);

        /**
        * Function that show the 3D components vector
        * Informations displayed are the "cfcSign" attributes (axis number and axis direction)
        */
        void showOuter(cfcStructTableType& outerComp);

        /**
        * Fonction permettant de visualiser le tableau de vecteurs contenant les composantes � 2 dimensions  
        * Funtion that show the array of 2D components vectors 
        * This function show the 2D components vector for all the 3D components and for the fixed dimension given as a parameter : middleComp[*][noFixed1]
        * Informations displayed are the "cfcSign" attributes (axis number and axis direction)
        * @param noFixed1 an integer value indicating from which fixed dimension the vector must be shown
        */
        void showMiddle(int noFixed1, cfcStructTableType& outerComp, middleCompType& middleComp, myGraphType& myGraph);

        /**
        * Fonction permettant de visualiser le tableau de vecteurs contenant les composantes � une dimension
        * Function that show the array of 1D components vectors 
        * This function show the 1D components vector for all the 3D components and all the 2D components for the fixed dimensions given as parameters : innerComp[*][noFixed1][*][noFixed2]
        * Informations displayed are the "cfcSign" attributes (axis number and axis direction) and the embeddedModuls attribute (moduls integrated in the component)
        * @param noFixed1 an integer value indicating from which first fixed dimension the vector must be shown
        * @param noFixed2 an integer value indicating from which second fixed dimension the vector must be shown
        * @param freeDim an integer value indicating the last dimension (not necessary because it does not index the array at all)
        */
        void showInner(int noFixed1, int noFixed2, int freeDim, cfcStructTableType& outerComp, middleCompType& middleComp, innerCompType& innerComp, myGraphType& myGraph);

        // -- Tool functions -- 

        /**
         * Function for getting the index number of the module which name is given as a parameter (tool function).
         * @param mod a string containing the modul name.
         * @param tab a string table containing all the modul names.
         * @param index an interger value indicating the range of tab parameter within the modul name is seeked 
         * @return an index on the modul number. 
         */
        int getModulIndex(string mod, const vector<string>& tab, int index);

        /**
        * Function that allow to see what is inside the orderTable given as a parameter
        * This function is just usefull for testing the result of the generation, so it is temporary 
        * @param theOrderTable a reference to a vector of Order objects 
        */
        void showTokens(Table<Order>& theOrderTable);

        /**
        * Check if a modul is integrated in an order statement
        * Function added since isCounterOrder() seems to have changed its behavior
        * @param nom the name of the modul that have to be seeked 
        * @param theOrderTable a reference on the order table in which the modul has to be seeked
        * @return a boolean value set to true if the module if the modul is integrated in an order statement
        */
        bool isInOrder(string nom, Table<Order>& theOrderTable);



 // ------------- FIN AJOUT CB GENERATION AUTOMATIQUE DE L'ORDER --------------- //

  private:

    /**
     * Allow to reverese the list or the root's son.
     * This list is very often of one simple element but sometime is more then one.
     * Here an example of more then one root's son:
     * order modinspace space
     *	order YA1 YA2
     *	  A B 
     *	forder
     *	order YA1 YA2
     *	  C D
     *	forder
     * forder.
     * The two order are directly connected to the root so they are the root's sons.
     */
    void reverseRootSon();	

    /**
     * Reverse all the token graph, it's a recursive method.
     * @param aToken a node of the graph.
     */
    void reverseTokenGraph(struct st_token *atoken);

    /**
     * Reverse the list of the sons of one father in the graph of tokens. 
     * To understand well it's important to remember the exect syntax of the order directive and to paint two or three examples of the graph.
     * The syntax of the order is:
     * ORD 		: 'order', Axe, List_AXE, Mod, List_inst , 'forder'
     * List_AXE	: vide | Axe, List_AXE
     * Axe		: YA1 | YA2 | YA3 | YB1 | YB2 | YB3
     * List_inst 	: vide | Mod, List_inst | ORD, list_inst
     * Mod		: doit exister parmi les inner de l'utilisateur 
     * @param aToken a node of the graph.
     */
    void reverseTokenList(struct st_token *atoken);

    /**
     * We go trough the new reversed graph to set the vector orderTokens.
     * @param aToken a node of the graph.
     * @param token the element to be changed of the orderTokens vector.
     */
    void setReversedVector(struct st_token *atoken, vector<string>::iterator *token);

    /**
     * Recursive method for the visualzation of the graph created in order to reverse the vector that contains the tokens.
     * This method is needed just for debug.
     * @param atoken the actual token we are visualizing.
     */
    void lirtokgr (struct st_token *atoken);

    struct st_token *rootToken;	    //!< The root tree. 

    Table<Modul> * theModulTable;   //!< A pointer to theModulTable in order to control in the backOrder() if we are managing a module token or not.
    // End of the reverse part--------------------------------------------------------------------------
    
    /*static Table<Modul> * theModTable;   //!< A pointer to theModulTable in order to control in the backOrder() if we are managing a module token or not.
    static const Table<Connection> * theConnectionTable; //! Pointer to the structure in the Translator.
    static Table<Space> * theSpaceTable; //! Pointer to the structure in the Translator.
    static Table<Operator> * theOperatorTable; //! Pointer to the structure in the Translator.
*/
    
    int orderPhase; //!< Phase syncronization in the order directive. The phases are: 1 modinspace, 2 spaceintraj, 3 end of all orders (a final block).
    string name;		    //!< The name of the Space in the case modinspace option otherwise the Trajectory name in the spaceintraj option.
  };

} // End namespace yao.
#endif //! YAO_ORDER_HPP_INCLUDED.
