//! \file    SpaceExtent.hpp
//! \brief   Declaration and implementation of the SpaceExtent class.
//! \version 2007/10/01 (yyyy/mm/dd) 
//! \author  Luigi Nardi <luiginardi(at)gmail.com> 
//
// This file is part of YAO, a framework for variational assimilation in
// numerical models.
// Copyright (c) 2001-onwards LOCEAN. All rights reserved.
//
// This program may be redistributed and/or modified under the terms of the GNU
// General Public License as published by the Free Software Foundation, either
// version 2 of the license, or (at your option) any later version.
//
// This program is provided AS IS and WITHOUT ANY WARRANTY, including any
// implied warranty of DESIGN, MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.


#ifndef YAO_SPACE_EXTENT_HPP_INCLUDED
#define YAO_SPACE_EXTENT_HPP_INCLUDED

#include <vector>
#include <iostream>

namespace yao{

  /**
   * SpaceExtent class stores the information of Yao space extension, so the shape.
   * The shape is represented with a number variable of integer in function of the
   * definition (i,j,k) of the Yao space. In three dimensions spaces the integer is three.
   */
  class SpaceExtent{

    public:

      SpaceExtent(){} //!< Default constructor, creates empty space extent.
      SpaceExtent(std::vector<int> aShape): theShape(aShape){} //!< Constructor.

      /**
       * Get the dimension, so the number of axes, of the space shape.
       * @return the dimension of the shape.
       */
      int getDimension() const { return theShape.size(); }

      /**
       * Get the shape of the space so a vector that contains from one to three integers in function to the dimension of the shape.
       * @return a constant shape vector.
       */
      const std::vector<int>& getShape() const { return theShape; }

      /**
       * Get the size YAx (x=1,2,3) of the space shape.
       * @param YAx is the number between 1 to 3 in order to identify one axe.
       * @return YAx (x=1,2,3).
       */
      int getYA(short YAx) const {
    	  if(YAx==1) return theShape.size()? theShape[0] : 0;
    	  if(YAx==2) return theShape.size() >= 2? theShape[1] : 0;
    	  if(YAx==3) return theShape.size() == 3? theShape[2] : 0;
    	  // Erroneous case
    	  // 0 is better than nothing..
    	  return 0;
      }
    
    protected:

      std::vector<int> theShape; //!< Shape (dimensions) of the space extent.
  };

} // Namespace yao
#endif // !YAO_SPACE_EXTENT_HPP_INCLUDED
